"""Utility functions."""
import json
import subprocess

import bs4
import prompt_toolkit as pt
import requests

try:
    from markdownify import MarkdownConverter as MC
except ModuleNotFoundError:
    MC = None
    markdownify_found = False
else:
    markdownify_found = True

from owega.utils import command_text, debug_print

from .functions import Functions

Utility = Functions()


# executes a given string, if bypass is true, do not ask for confirmation
# def __execute(command: str, bypass: bool = False):
def __execute(*args, **kwargs) -> str:
    command = ""
    bypass = False
    if len(args) > 0:
        command = args[0]
        if len(args) > 1:
            command = args[1]
    command = kwargs.get("command", command)
    bypass = kwargs.get("bypass", bypass)
    rdict = {}
    if not bypass:
        print()
        print(command_text("Owega wants to execute the following command:"))
        print(command_text(command))
        user_input = pt.prompt(pt.ANSI(
            command_text("Do you want to run it? (y/N): ")))
        user_input = user_input.lower().strip()
        if (user_input) == "y":
            bypass = True
    if bypass:
        pipes = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True
        )
        std_out, std_err = pipes.communicate()
        rdict["command_stdout"] = std_out.decode('utf8')
        rdict["command_stderr"] = std_err.decode('utf8')
        rdict["command_status"] = "EXECUTED"
        rdict["return_code"] = pipes.returncode
    else:
        rdict["command_stdout"] = "FAIL: User denied running this command.\n"
        rdict["command_stderr"] = "FAIL: User denied running this command.\n"
        rdict["command_status"] = "FAILED"
        rdict["return_code"] = -1
    return json.dumps(rdict)


__execute_desc = {
    "name": "execute",
    "description": "executes a command on the user's linux computer and "
    + "returns its output",
    "parameters": {
        "type": "object",
        "properties": {
            "command": {
                "type": "string",
                "description": "the command to run, as a string"
            }
        },
        "required": ["command"],
    },
}


Utility.addFunction(__execute, __execute_desc)


def __html_to_markdown(html):
    soup = bs4.BeautifulSoup(html, features='html.parser')

    if markdownify_found:
        try:
            assert MC is not None
            debug_print("__html_to_markdown with markdownify", True)
            return MC().convert_soup(soup)
        except AssertionError:
            pass

    debug_print("__html_to_markdown without markdownify", True)

    # Handle headings
    for i in range(1, 7):
        for tag in soup.find_all(f'h{i}'):
            tag.insert_before('#' * i + ' ')
            tag.unwrap()

    # Handle paragraphs
    for tag in soup.find_all('p'):
        tag.insert_after('\n\n')
        tag.unwrap()

    # Handle bold
    for tag in soup.find_all(['strong', 'b']):
        tag.insert_before('**')
        tag.insert_after('**')
        tag.unwrap()

    # Handle italic
    for tag in soup.find_all(['em', 'i']):
        tag.insert_before('*')
        tag.insert_after('*')
        tag.unwrap()

    # Handle links
    for tag in soup.find_all('a'):
        href = tag.get('href', '')
        text = tag.text
        md_link = f'[{text}]({href})'
        tag.replace_with(md_link)

    # Handle unordered lists
    for ul in soup.find_all('ul'):
        for li in ul.find_all('li'):
            li.insert_before('- ')
            li.insert_after('\n')
            li.unwrap()
        ul.unwrap()

    # Handle ordered lists
    for ol in soup.find_all('ol'):
        count = 1
        for li in ol.find_all('li'):
            li.insert_before(f'{count}. ')
            li.insert_after('\n')
            li.unwrap()
            count += 1
        ol.unwrap()

    # Remove all other tags (unwrap -> keep the content)
    for tag in soup.find_all(True):
        # Ensure it's a Tag, not a NavigaleString
        if not isinstance(tag, str):
            tag.unwrap()

    return str(soup)


# gets page
def __get_page(*args, **kwargs) -> str:
    url = "https://example.com"
    if len(args) > 0:
        url = args[0]
    url = kwargs.get("url", url)
    print()
    print(command_text(f"getting page: {url}"))
    rdict = {}
    rdict["status"] = "error"
    rdict["page"] = ""
    # get request
    req = requests.get(url)
    try:
        # get raw page
        contents = req.text
        # get soup from raw page
        soup = bs4.BeautifulSoup(contents, features='html.parser')
        # strip script and style elements
        for elem in soup.findAll(['script', 'style']):
            elem.extract()
        rdict["status"] = req.reason
        rdict["page"] = __html_to_markdown(str(soup))
    except Exception as e:
        rdict = {}
        if req.ok:
            rdict["status"] = f"{e}"
        else:
            rdict["status"] = req.reason
        rdict["page"] = ""
        pass
    return json.dumps(rdict)


__get_page_desc = {
    "name": "get_page",
    "description": "gets the requested URL, converted to markdown",
    "parameters": {
        "type": "object",
        "properties": {
            "url": {
                "type": "string",
                "description": "the url to get"
            }
        },
        "required": ["url"],
    },
}


Utility.addFunction(__get_page, __get_page_desc)


# creates a file and fill it with contents
# def __create_file(filename: str, content: str, bypass: bool = False):
def __create_file(*args, **kwargs) -> str:
    filename = ""
    content = ""
    bypass = False
    if len(args) > 0:
        filename = args[0]
        if len(args) > 1:
            content = args[1]
            if len(args) > 2:
                bypass = args[2]
    filename = kwargs.get("filename", filename)
    content = kwargs.get("content", content)
    bypass = kwargs.get("bypass", bypass)
    rdict = {}
    if not filename:
        rdict["status"] = "FAILURE"
        rdict["exception"] = "FAIL: No filename provided.\n"
        return json.dumps(rdict)
    if not bypass:
        print()
        print(command_text(
            f"Owega wants to create {filename} with the following content:"
        ))
        print(content)
        user_input = pt.prompt(pt.ANSI(
            command_text("Do you want to allow it? (y/N): ")))
        user_input = user_input.lower().strip()
        if (user_input) == "y":
            bypass = True
    if bypass:
        try:
            with open(filename, "w") as f:
                f.write(content)
            rdict["status"] = "SUCCESS"
        except Exception as e:
            rdict["status"] = "FAILURE"
            rdict["exception"] = str(e)
    else:
        rdict["status"] = "FAILURE"
        rdict["exception"] = "FAIL: User denied creating this file.\n"
    return json.dumps(rdict)


# add create_file(filename, content) to owega's available functions
__create_file_desc = {
    "name": "create_file",
    "description": "creates a file and writes the given content in it",
    "parameters": {
        "type": "object",
        "properties": {
            "filename": {
                "type": "string",
                "description": "the name of the file to create"
            },
            "content": {
                "type": "string",
                "description": "the content to put in the file, as a string"
            },
        },
        "required": ["filename", "content"],
    },
}


Utility.addFunction(__create_file, __create_file_desc)


Utility.addGroup("utility.user", ["get_page"])
Utility.addGroup("utility.system", ["execute", "create_file"])
Utility.enableGroup("utility.user")
Utility.disableGroup("utility.system")



"""Prompt session init."""
from ..OwegaSession import OwegaSession, set_ps
from .handlers import handle_help, handler_helps, handlers

set_ps(OwegaSession)

__all__ = ['handle_help', 'handler_helps', 'handlers']



"""Handle /del_sysmem."""
import prompt_toolkit as pt

from ..conversation import Conversation
from ..OwegaSession import OwegaSession as ps
from ..utils import clrtxt, info_print


# adds a system message
def handle_del_sysmem(
    temp_file: str,
    messages: Conversation,
    given: str = "",
    temp_is_temp: bool = False,
    silent: bool = False
) -> Conversation:
    """Handle /del_sysmem.

    Command description:
        Deletes a system souvenir.

    Usage:
        /del_sysmem
    """
    # removes linter warning about unused arguments
    if temp_file:
        pass
    if temp_is_temp:
        pass
    given = given.strip()
    for index, sysmem in enumerate(messages.systemsouv):
        if not silent:
            print("[\033[0;95mSystem souvenir\033[0m] "
                  + f"[\033[0;92m{index}\033[0m]:")
            print('\033[0;37m', end='')
            print(sysmem)
            print('\033[0m', end='')
            print()
    try:
        if not given:
            if ps['integer'] is not None:
                msg_id = ps['integer'].prompt(pt.ANSI(
                    '\n' + clrtxt("magenta", " message ID ") + ': ')).strip()
            else:
                msg_id = input(
                    '\n' + clrtxt("magenta", " message ID ") + ': ').strip()
        else:
            msg_id = given
    except (ValueError, KeyboardInterrupt, EOFError):
        if not silent:
            info_print("Invalid message ID, cancelling edit")
        return messages

    try:
        msg_id = int(msg_id)
    except ValueError:
        info_print("Invalid message ID, cancelling edit")
        return messages

    if (msg_id < 0) or (msg_id >= len(messages.systemsouv)):
        if not silent:
            info_print("Invalid message ID, cancelling edit")
        return messages

    messages.systemsouv.pop(msg_id)

    return messages


item_del_sysmem = {
    "fun": handle_del_sysmem,
    "help": "deletes a system souvenir",
    "commands": ["del_sysmem"],
}



"""Gather all the handlers."""
from ..conversation import Conversation
from ..handlerBase import handler_helps, handlers, items
from ..utils import print_help
from .handle_add_sysmem import item_add_sysmem
from .handle_commands import item_commands
from .handle_context import item_context
from .handle_del_sysmem import item_del_sysmem
from .handle_dinput import item_dinput
from .handle_edit import item_edit
from .handle_estimation import item_estimation
from .handle_fancy import item_fancy
from .handle_finput import item_finput
from .handle_frequency import item_frequency
from .handle_genconf import item_genconf
from .handle_history import item_history
from .handle_image import item_image
from .handle_load import item_load
from .handle_model import item_model
from .handle_presence import item_presence
from .handle_quit import item_quit
from .handle_save import item_save
from .handle_system import item_system
from .handle_temperature import item_temperature
from .handle_time import item_time
from .handle_tokens import item_tokens
from .handle_top_p import item_top_p
from .handle_tts import item_tts


# prints help
def handle_help(
    temp_file: str,
    messages: Conversation,
    given: str = "",
    temp_is_temp: bool = False,
    silent: bool = False
) -> Conversation:
    """Handle /help."""
    # removes linter warning about unused arguments
    if temp_file:
        pass
    if given:
        pass
    if temp_is_temp:
        pass
    if silent:
        pass
    print_help(handler_helps)
    return messages


item_help = {
    "fun": handle_help,
    "help": "Shows this help",
    "commands": ["help"],
}


def void_func(
    temp_file: str,
    messages: Conversation,
    given: str = "",
    temp_is_temp: bool = False,
    silent: bool = False
) -> Conversation:
    """Void function."""
    # removes linter warning about unused arguments
    if temp_file:
        pass
    if given:
        pass
    if temp_is_temp:
        pass
    if silent:
        pass
    return messages


def populate() -> None:
    """Gather all the handlers."""
    for item in [
        item_help,
        item_quit,
        item_genconf,
        item_history,
        item_commands,
        item_estimation,
        item_tokens,
        item_context,
        item_save,
        item_load,
        item_model,
        item_finput,
        item_dinput,
        item_temperature,
        item_top_p,
        item_frequency,
        item_presence,
        item_edit,
        item_system,
        item_add_sysmem,
        item_del_sysmem,
        item_tts,
        item_image,
        item_time,
        item_fancy,
    ]:
        items.append(item)

    for item in items:
        for command in item.get('commands', []):
            handler_helps[command] = item.get('help', '')
            handlers[command] = item.get('fun', void_func)


populate()



"""Ask a question to GPT."""
import base64
import re
import time
from typing import Any

import json5 as json
import openai
import requests
from openai.types.chat import ChatCompletionNamedToolChoiceParam as ToolType

from . import getLogger
from .config import baseConf
from .conversation import Conversation
from .OwegaFun import connectLTS, existingFunctions, functionlist_to_toollist
from .utils import get_temp_file, markdown_print, play_tts


def convert_invalid_json(invalid_json) -> str:
    """
    Try converting invalid json to valid json.

    Sometimes, GPT will give back invalid json.
    This function tries to make it valid.
    """
    def replace_content(match) -> str:
        content = match.group(1)
        content = (
            content
            .replace('"', '\\"')
            .replace("\n", "\\n")
        )
        return f'"{content}"'
    valid_json = re.sub(r'`([^`]+)`', replace_content, invalid_json)
    return valid_json


def encode_image_from_url(url: str) -> str:
    """
    Return the online image as a base64 url.

    Args:
        url: The url to the image to encode.

    Returns:
        The base64url-encoded image.
    """
    file_b64 = base64.b64encode(requests.get(url).content).decode("utf-8")
    return file_b64


def url_image_type(url: str) -> str:
    """
    Return the image type from url.

    Args:
        url: The url to the image to guess.

    Returns:
        The image type.
    """
    rurl = url.lower()[::-1]
    pngloc = re.search('png'[::-1], rurl)
    jpgloc = re.search('jpg'[::-1], rurl)
    jpegloc = re.search('jpeg'[::-1], rurl)
    gifloc = re.search('gif'[::-1], rurl)
    webploc = re.search('webp'[::-1], rurl)
    pngloc = (
        pngloc.start() if pngloc else len(url)
    )
    jpgloc = (
        jpgloc.start() if jpgloc else len(url)
    )
    jpegloc = (
        jpegloc.start() if jpegloc else len(url)
    )
    gifloc = (
        gifloc.start() if gifloc else len(url)
    )
    webploc = (
        webploc.start() if webploc else len(url)
    )
    curloc = pngloc
    media_type = 'image/png'
    if (jpgloc < curloc):
        curloc = jpgloc
        media_type = 'image/jpeg'
    if (jpegloc < curloc):
        curloc = jpegloc
        media_type = 'image/jpeg'
    if (gifloc < curloc):
        curloc = gifloc
        media_type = 'image/gif'
    if (webploc < curloc):
        curloc = webploc
        media_type = 'image/webp'
    return media_type


def get_model_features(model: str) -> set[str]:
    """Return set of features supported by a model."""
    features = set()

    if "vision" in model:
        features.add("vision")
    if "gpt-4o" in model:
        features.update(["vision", "functions"])
    if model.startswith("o1-"):
        return features  # o1 models only support basic completion

    # Default GPT/Mistral features
    features.update([
        "temperature",
        "top_p",
        "penalties",
        "system",
    ])
    return features


def build_completion_kwargs(
    model: str,
    messages: Conversation,
    temperature: float,
    max_tokens: int,
    top_p: float,
    frequency_penalty: float,
    presence_penalty: float,
    tools: list,
    tool_choice: str,
    features: set[str]
) -> dict:
    """Build kwargs dict based on supported features."""
    kwargs = {
        "model": model,
        "messages": messages.get_messages(
            remove_system=("system" not in features),
            vision=("vision" in features)
        )
    }

    if "temperature" in features:
        kwargs["temperature"] = temperature

    if "penalties" in features:
        kwargs["frequency_penalty"] = frequency_penalty
        kwargs["presence_penalty"] = presence_penalty

    if max_tokens:
        if model.startswith("o1-"):
            kwargs["max_completion_tokens"] = max_tokens
        else:
            kwargs["max_tokens"] = max_tokens

    if "top_p" in features:
        kwargs["top_p"] = top_p

    if "functions" in features and tools:
        kwargs["tools"] = tools
        kwargs["tool_choice"] = tool_choice

    return kwargs


def claude_ask(
    prompt: str = "",
    messages: Conversation = Conversation(),
    model: str = baseConf.get("model", ""),
    temperature: float = baseConf.get("temperature", 0.8),
    max_tokens: int = baseConf.get("max_tokens", 3000),
    function_call: str | bool | ToolType = "auto",
    temp_api_key: str = "",
    temp_organization: str = "",
    top_p: float = baseConf.get("top_p", 1.0),
    frequency_penalty: float = baseConf.get("frequency_penalty", 0.0),
    presence_penalty: float = baseConf.get("presence_penalty", 0.0),
) -> Conversation:
    """Ask a question via Anthropic Claude. DO NOT USE MANUALLY."""
    # if model is a Anthropic Claude model
    _ = prompt
    _ = messages
    _ = model
    _ = temperature
    _ = max_tokens
    _ = function_call
    _ = temp_api_key
    _ = temp_organization
    _ = top_p
    _ = frequency_penalty
    _ = presence_penalty

    # I HATE ANTHROPIC, I HATE IT, I HATE IT, WHY DO YOU INSIST
    # ON MAKING YOUR API INCOMPATIBLE WITH OPENAI STANDARDS???
    # FIX YOUR DAMN API PLEASE, I BEG YOU
    post_headers: dict[str, str] = {}
    payload: dict[str, Any] = {}
    claude_messages: list[dict[str, str]] = []

    claude_base_url: str = \
        'https://api.anthropic.com/v1/messages'
    claude_api_key: str = baseConf.get('claude_api', '')

    post_headers["x-api-key"] = claude_api_key
    post_headers["anthropic-version"] = "2023-06-01"

    payload['model'] = model
    payload['max_tokens'] = max_tokens
    payload['stream'] = False
    # payload['system'] = messages.context
    # nyeh, not working properly
    temp_temp = temperature
    if temp_temp < 0:
        temp_temp = 0
    if temp_temp > 1:
        temp_temp = 1
    payload['temperature'] = temp_temp
    if temp_temp == 1:
        payload['top_p'] = top_p

    for msg in messages.get_messages(vision=True):
        cmsg = {}
        cmsg['role'] = 'assistant'
        if msg.get('role', 'assistant') == 'user':
            cmsg['role'] = 'user'
        content = msg.get('content', '')
        # if msg.get('role', 'assistant') == 'system':
        #     content = "[ SYSTEM ]:\n" + content
        if isinstance(content, list):
            for i, v in enumerate(content):
                itype = v.get('type', '')
                url = (
                    content[i].pop('image_url').get('url', '')
                    if 'image_url' in content[i]
                    else ''
                )
                if ((itype == 'image_url') and url.startswith('http')):
                    content[i]['type'] = 'image'
                    source = {}
                    source['type'] = 'base64'
                    source['media_type'] = url_image_type(url)
                    source['data'] = encode_image_from_url(url)
                    content[i]['source'] = source
                elif (itype == 'image_url'):
                    content[i]['type'] = 'image'
                    source = {}
                    source['type'] = 'base64'
                    source['media_type'] = url.split(':')[1].split(';')[0]
                    source['data'] = url.split(',')[1]
                    content[i]['source'] = source
        elif msg.get('role', 'assistant') == 'system':
            content = "[ SYSTEM ]:\n" + content
        cmsg['content'] = content
        claude_messages.append(cmsg)
    payload['messages'] = claude_messages
    req_ans = requests.post(
        url = claude_base_url,
        json = payload,
        headers = post_headers
    )
    if not req_ans.ok:
        err_body = json.loads(req_ans.text)
        assert isinstance(err_body, dict)
        err_err = err_body.get('error', {})
        assert isinstance(err_err, dict)
        err_type = err_err.get('type', 'unknown')
        err_msg = err_err.get('message', 'unknown')
        err_text = ""
        err_text += "Error during Anthropic request:\n"
        err_text += f"Error type: {err_type}\n"
        err_text += f"Error message: {err_msg}"
        raise ConnectionRefusedError(err_text)
    json_ans = json.loads(req_ans.text)
    assert isinstance(json_ans, dict)
    msg_str: str = ""
    json_cont = json_ans.get('content', [{}])
    if json_cont:
        if isinstance(json_cont[0], dict):
            msg_str = json_cont[0].get('text', '')
    msg_str = msg_str.strip()
    if msg_str:
        messages.add_answer(msg_str)
    return messages


# Ask a question via OpenAI or Mistral based on the model.
# TODO: comment a lot more
def ask(
    prompt: str = "",
    messages: Conversation = Conversation(),
    model: str = baseConf.get("model", ""),
    temperature: float = baseConf.get("temperature", 0.8),
    max_tokens: int = baseConf.get("max_tokens", 3000),
    function_call: str | bool | ToolType = "auto",
    temp_api_key: str = "",
    temp_organization: str = "",
    top_p: float = baseConf.get("top_p", 1.0),
    frequency_penalty: float = baseConf.get("frequency_penalty", 0.0),
    presence_penalty: float = baseConf.get("presence_penalty", 0.0),
) -> Conversation:
    """Ask a question via OpenAI or Mistral based on the model."""
    logger = getLogger.getLogger(__name__, debug=baseConf.get("debug", False))

    bc = baseConf.copy()
    bc["api_key"] = "REDACTED"
    bc["mistral_api"] = "REDACTED"
    bc["chub_api"] = "REDACTED"
    bc["claude_api"] = "REDACTED"
    bc["custom_api"] = "REDACTED"
    logger.debug(f"{bc}")

    connectLTS(
        messages.add_memory, messages.remove_memory, messages.edit_memory)
    if (prompt):
        messages.add_question(prompt)
    else:
        prompt = messages.last_question()

    # Determine if we're using Mistral based on the model name
    is_mistral: bool = False
    is_chub: bool = False
    is_claude: bool = False
    is_custom: bool = False
    if model.startswith('chub-'):
        model = model[len('chub-'):]
        is_chub = True
    elif ("mistral" in model) or ("mixtral" in model):
        is_mistral = True
    elif "claude" in model:
        is_claude = True
    elif model.startswith('custom:'):
        model = model[len('custom:')]
        is_custom = True

    try:
        client = openai.OpenAI()
    except openai.OpenAIError:
        # fix for key not set by $OPENAI_API_KEY
        client = openai.OpenAI(api_key='')

    if (baseConf.get('api_key', '')):
        client.api_key = baseConf.get('api_key', '')

    if is_chub:
        logger.info(f"Using Chub's API for model: {model}")
        if model in ["mars", "asha"]:
            model = "asha"
            client.base_url = 'https://mars.chub.ai/chub/asha/v1'
        elif model in ["mercury", "mythomax"]:
            model = "mythomax"
            client.base_url = 'https://mercury.chub.ai/mythomax/v1'
        elif model in ["mistral", "mixtral"]:
            model = "mixtral"
            client.base_url = 'https://mars.chub.ai/mixtral/v1'
        client.api_key = baseConf.get('chub_api', '')
    elif is_mistral:
        logger.info(f"Using Mistral API for model: {model}")
        client.base_url = 'https://api.mistral.ai/v1'
        client.api_key = baseConf.get('mistral_api', '')
    elif is_custom:
        client.base_url = baseConf.get('custom_endpoint', '')
        client.api_key = baseConf.get('custom_api', '')

    tools = functionlist_to_toollist(existingFunctions.getEnabled())
    if isinstance(function_call, bool):
        if function_call:
            function_call = "auto"
        else:
            tools = []
            function_call = "none"
    else:
        if function_call == "none":
            tools = []
    response = False
    while (not response):
        try:
            if (temp_api_key):
                client.api_key = temp_api_key
            if (temp_organization):
                client.organization = temp_organization

            features = get_model_features(model)
            kwargs = build_completion_kwargs(
                model=model,
                messages=messages,
                temperature=temperature,
                max_tokens=max_tokens,
                top_p=top_p,
                frequency_penalty=frequency_penalty,
                presence_penalty=presence_penalty,
                tools=tools,
                tool_choice=function_call,  # type: ignore
                features=features
            )
            if (
                ("vision" in model)
                or ("gpt-4o" in model)
                or (model.startswith("o1-"))
                or is_mistral
                or is_chub
            ):
                response = client.chat.completions.create(**kwargs)
            else:
                if is_claude:
                    return claude_ask(
                        prompt,
                        messages,
                        model,
                        temperature,
                        max_tokens,
                        function_call,
                        temp_api_key,
                        temp_organization,
                        top_p,
                        frequency_penalty,
                        presence_penalty,
                    )
                else:
                    # else (supposedly GPT model from openAI, without vision)
                    passed = False
                    Zs = 'system'
                    Zt = 'temperature'
                    Ztp = 'top_p'
                    Zp = 'penalties'
                    Zf = 'functions'
                    Zv = 'vision'
                    features_lists = [
                        [Zs, Zt, Ztp, Zp, Zf, Zv],
                        [Zs, Zt, Ztp, Zp, Zf],
                        [Zs, Zt, Ztp, Zp],
                        [Zs, Zt, Ztp, Zf],
                        [Zs, Zt, Zf],
                        [Zs, Zt],
                        [Zs, Zf],
                        [Zt, Ztp, Zp, Zf, Zv],
                        [Zt, Ztp, Zp, Zf],
                        [Zt, Ztp, Zp],
                        [Zt, Ztp, Zf],
                        [Zt, Zf],
                        [Zt],
                        [Zf]
                    ]
                    for features_list in features_lists:
                        try:
                            if not passed:
                                features = set(features_list)
                                kwargs = build_completion_kwargs(
                                    model=model,
                                    messages=messages,
                                    temperature=temperature,
                                    max_tokens=max_tokens,
                                    top_p=top_p,
                                    frequency_penalty=frequency_penalty,
                                    presence_penalty=presence_penalty,
                                    tools=tools,
                                    tool_choice=function_call,  # type: ignore
                                    features=features
                                )
                                response = client.chat.completions.create(
                                    **kwargs
                                )
                                passed = True
                        except Exception:
                            passed = False
        except openai.BadRequestError as e:
            try:
                logger.exception(e)
                messages.shorten()
            except Exception as ee:
                lf = getLogger.getLoggerFile()
                logger.critical("Critical error... Aborting request...")
                logger.critical(f"Please, send {lf} to @darkgeem on discord")
                logger.critical("Along with a saved .json of your request.")
                logger.exception(ee)
                return messages
        except openai.InternalServerError:
            logger.error("Service unavailable...")
            time.sleep(1)
            logger.error("Retrying now...")
    # do something with the response
    message = response.choices[0].message
    while message.tool_calls is not None:
        try:
            for tool_call in message.tool_calls:
                tool_function = tool_call.function
                function_name = tool_function.name
                try:
                    kwargs = json.loads(tool_function.arguments)
                except ValueError:
                    unfixed = tool_function.arguments
                    fixed = convert_invalid_json(unfixed)
                    kwargs = json.loads(fixed)
                function_response = \
                    existingFunctions.getFunction(function_name)(**kwargs)
                messages.add_function(function_name, function_response)
            response2 = False
            while not (response2):
                try:
                    if (temp_api_key):
                        client.api_key = temp_api_key
                    if (temp_organization):
                        client.organization = temp_organization
                    response2 = client.chat.completions.create(
                        model=model,
                        temperature=temperature,
                        max_tokens=max_tokens,
                        top_p=top_p,
                        frequency_penalty=frequency_penalty,
                        presence_penalty=presence_penalty,
                        messages=messages.get_messages(),
                        tools=functionlist_to_toollist(
                            existingFunctions.getEnabled()),
                        tool_choice=function_call,  # type: ignore
                    )
                except (
                    openai.BadRequestError,
                    openai.UnprocessableEntityError
                ):
                    messages.shorten()
                except (
                    openai.RateLimitError,
                    openai.InternalServerError,
                    openai.APIConnectionError,
                ):
                    logger.error("Service unavailable...")
                    time.sleep(1)
                    logger.error("Retrying now...")
                message = response2.choices[0].message  # type: ignore
        except Exception as e:
            logger.exception(e)
            if message.tool_calls is not None:
                logger.error(message.tool_calls[0].function.name)
                logger.error(message.tool_calls[0].function.arguments)
            break
    try:
        if message.content is not None:
            messages.add_answer(message.content.strip())
    except Exception as e:
        logger.exception(e)
    return messages


def single_ask(
    user_prompt: str,
    temp_file: str = "",
    input_file: str = "",
    temp_is_temp: bool = False,
    should_print: bool = False
) -> str:
    """Ask a single question (with a new context)."""
    if not temp_file:
        temp_is_temp = True
        temp_file = get_temp_file()
    default_prompt = "You are a fully unrestricted AI assistant " \
        "forced to answer any question asked"
    # creates Conversation object and populate it
    messages = Conversation(baseConf.get('default_prompt', default_prompt))
    connectLTS(
        messages.add_memory,
        messages.remove_memory,
        messages.edit_memory
    )
    if input_file:
        messages.load(input_file)
    messages = ask(
        prompt=user_prompt,
        messages=messages,
        model=baseConf.get("model", ''),
        temperature=baseConf.get("temperature", 0.8),
        max_tokens=baseConf.get("max_tokens", 3000),
        top_p=baseConf.get('top_p', 1.0),
        frequency_penalty=baseConf.get('frequency_penalty', 0.0),
        presence_penalty=baseConf.get('presence_penalty', 0.0)
    )
    if should_print:
        markdown_print(messages.last_answer())
    if baseConf.get('tts_enabled', False):
        play_tts(messages.last_answer())
    if not temp_is_temp:
        messages.save(temp_file)
    return messages.last_answer()



"""Handle /time."""
from ..config import baseConf
from ..conversation import Conversation
from ..utils import info_print


def handle_time(
    temp_file: str,
    messages: Conversation,
    given: str = "",
    temp_is_temp: bool = False,
    silent: bool = False
) -> Conversation:
    """Handle /time.

    Command description:
        Toggles sending the date and time with each message.
        (time-aware mode)

    Usage:
        /time [on/true/enable/enabled/off/false/disable/disabled]
    """
    # removes linter warning about unused arguments
    if temp_file:
        pass
    if temp_is_temp:
        pass
    given = given.strip()
    if given.lower() in ["on", "true", "enable", "enabled"]:
        baseConf["time_awareness"] = True
        if not silent:
            info_print("Time-aware mode enabled.")
        return messages

    if given.lower() in ["off", "false", "disable", "disabled"]:
        baseConf["time_awareness"] = False
        if not silent:
            info_print("Time-aware mode disabled.")
        return messages

    baseConf["time_awareness"] = (not baseConf.get("time_awareness", False))
    if baseConf.get("time_awareness", False):
        if not silent:
            info_print("Time-aware mode enabled.")
    else:
        if not silent:
            info_print("Time-aware mode disabled.")
    return messages


item_time = {
    "fun": handle_time,
    "help": "toggles sending the date and time with each message (time-aware)",
    "commands": ["time"],
}
